﻿// .Net Follower http://dotnetfollower.com

using System;
using System.Xml;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace Upgrade2007BdcModelTo2010
{
    public class Sp07BdcModelToSp10Converter
    {
        #region auxiliary types
        protected class NodeMatch
        {
            public string WholeNode;
            public string Content;
            public string Attributes;
            public string WholeOpeningTag;
            public string TagName;            
        }

        protected class ReplaceNodeMatch
        {
            public NodeMatch FoundNodeMatch;
            public Match     OriginMatch;
            public bool      Skip;
        }
        #endregion

        protected static string _attrExtractorTemplate = @"{0}\s*=\s*(""(?<val>.*?)""|'(?<val>.*?)'|(?<val>[^'"">\s]))+";
        // the next regex expression works only for tags that don't contain the similar tags inside
        protected static string _tagExtractorTemplate  = @"(?<openTag><\s*(?<tagName>{0})(?<attr>((\s+(\w|:)+(\s*=\s*(?:"".*?""|'.*?'|[^'"">\s]+))?)+\s*|\s*)?)>)(?<content>(.|\n)*?)<\s*/\s*{0}\s*>|(?<openTag><\s*(?<tagName>{0})(?<attr>((\s+(\w|:)+(\s*=\s*(?:"".*?""|'.*?'|[^'"">\s]+))?)+)?)\s*/\s*>)";
        protected static Regex  _typeExtractor         = new Regex(@"(?<type>[\w.]+)\s*,\s*(?<dll>.+)\s*,\s*Version\s*=\s*(?<version>[\d.]+)\s*,\s*Culture\s*=\s*(?<culture>\w+)\s*,\s*PublicKeyToken\s*=\s*(?<PublicKeyToken>\w+)", RegexOptions.IgnoreCase | RegexOptions.Compiled);       
        
        protected static Dictionary<string, string> _typesToReplace = new Dictionary<string,string>()
        {
            { "Microsoft.Office.Server.ApplicationRegistry.SystemSpecific.WebService.HttpAuthenticationMode", "System.String" },
            { "Microsoft.Office.Server.ApplicationRegistry.SystemSpecific.Db.DbAuthenticationMode",           "System.String" },
            { "Microsoft.Office.Server.ApplicationRegistry.SystemSpecific.Db.DbAccessProvider",               "System.String" },

            { "Microsoft.SharePoint.Portal.SingleSignon.SpsSsoProvider, Microsoft.SharePoint.Portal.SingleSignon, Version=12.0.0.0, Culture=neutral, PublicKeyToken=71e9bce111e9429c", "Microsoft.Office.SecureStoreService.Server.SecureStoreProvider, Microsoft.Office.SecureStoreService, Version=14.0.0.0, Culture=neutral, PublicKeyToken=71e9bce111e9429c" }
        };

        protected string _lobDateTimeMode  = "UTC";
        protected string _lobSystemVersion;
        protected string _lobSystemName;

        public string Convert(string source)
        {
            string res = source;            

            // replace banned types with allowed ones
            res = ReplaceBannedTypes(res);

            // add outer Model tag
            res = AddOuterModel(res);

            // adjust Entity tags
            res = AdjustEntityTags(res);

            // adjust MethodInstance tags
            res = AdjustMethodInstanceTags(res);

            // adjust TypeDescriptor tags
            res = AdjustTypeDescriptorTags(res);

            // return indented xml-based string
            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.LoadXml(res);
            return XmlDocToString(xmlDoc);            
        }        

        protected static string ReplaceBannedTypes(string source)
        {            
            StringBuilder sb = new StringBuilder(source);

            foreach(KeyValuePair<string, string> type in _typesToReplace)
                sb.Replace(type.Key, type.Value);

            return sb.ToString();
        }

        protected string AddOuterModel(string source)
        {
            string res = ReplaceNode("lobSystem", source, delegate(ReplaceNodeMatch replaceNodeMatch)
            {
                // LobSystem attributes
                string attrs = replaceNodeMatch.FoundNodeMatch.Attributes;

                // remove namespace attributes
                attrs = RemoveAttr("(xmlns:xsi|xsi:schemaLocation|xmlns)", attrs);

                // get version attribute value and remove version attribute                
                attrs = RemoveAttr("version", attrs, out _lobSystemVersion);

                // find the name attribute
                _lobSystemName = GetAttrVal("name", attrs);                

                // get LobSystem tag comment
                string tagContent = replaceNodeMatch.FoundNodeMatch.Content.Trim();

                replaceNodeMatch.FoundNodeMatch.Attributes = attrs;
                replaceNodeMatch.FoundNodeMatch.Content    = tagContent;

                // create new LobSystem tag
                string newLobSystem = FormNode(replaceNodeMatch.FoundNodeMatch);
                
                // create replacement
                string replacement = string.Format(
                        @"<Model xmlns:xsi=""http://www.w3.org/2001/XMLSchema-instance"" 
                        xmlns=""http://schemas.microsoft.com/windows/2007/BusinessDataCatalog"" 
                        xsi:schemaLocation=""http://schemas.microsoft.com/windows/2007/BusinessDataCatalog http://schemas.microsoft.com/windows/2007/BusinessDataCatalog/BDCMetadata.xsd"" 
                        Name=""{0} Model"">
                        <LobSystems>
                        {1}
                        </LobSystems>
                        </Model>", _lobSystemName, newLobSystem);

                return replacement;
            });

            return res;
        }

        protected string AdjustEntityTags(string source)
        {
            string res = ReplaceNode("entity", source, delegate(ReplaceNodeMatch replaceNodeMatch)
            {
                // entity attributes
                string attrs = replaceNodeMatch.FoundNodeMatch.Attributes;

                attrs = AddAttr("Namespace", _lobSystemName,    attrs);
                attrs = AddAttr("Version",   _lobSystemVersion, attrs);
                
                // get Entity tag comment
                string tagContent = replaceNodeMatch.FoundNodeMatch.Content.Trim();

                // remove an assembly name in Identifier-elements
                tagContent = ReplaceNode("identifier", tagContent, delegate(ReplaceNodeMatch identReplaceNodeMatch)
                {
                    string identAttrs = identReplaceNodeMatch.FoundNodeMatch.Attributes;

                    string typeNameAttrVal = GetAttrVal("TypeName", identAttrs);
                    Match typeNameAttrValMatch = _typeExtractor.Match(typeNameAttrVal);
                    if (!typeNameAttrValMatch.Success)
                    {
                        identReplaceNodeMatch.Skip = true;
                        return null;
                    }

                    identAttrs = RemoveAttr("TypeName", identAttrs);
                    identAttrs = AddAttr("TypeName", typeNameAttrValMatch.Groups["type"].Value, identAttrs);

                    identReplaceNodeMatch.FoundNodeMatch.Attributes = identAttrs;
                    return FormNode(identReplaceNodeMatch.FoundNodeMatch);
                });

                replaceNodeMatch.FoundNodeMatch.Attributes = attrs;
                replaceNodeMatch.FoundNodeMatch.Content    = tagContent;

                return FormNode(replaceNodeMatch.FoundNodeMatch);
            });

            return res;
        }

        protected string AdjustMethodInstanceTags(string source)
        {
            List<NodeMatch> methodNodes = GetNodes("Method", source);

            StringBuilder sb = new StringBuilder(source);
            foreach (NodeMatch nodeMatch in methodNodes)
            {
                List<NodeMatch> methodInstanceNodes = GetNodes("methodinstance", nodeMatch.Content);

                bool specificFinderFound = false;
                foreach (NodeMatch methodInstanceNodeMatch in methodInstanceNodes)
                {
                    string attrs = methodInstanceNodeMatch.Attributes;
                    attrs = RemoveAttr("ReturnTypeDescriptorLevel", attrs);

                    string returnTypeDescriptorNameVal = null;
                    attrs = RemoveAttr("ReturnTypeDescriptorName", attrs, out returnTypeDescriptorNameVal);

                    string methodInstanceType = GetAttrVal("Type", methodInstanceNodeMatch.Attributes);
                    bool isSpecificFinder = "SpecificFinder".Equals(methodInstanceType, StringComparison.OrdinalIgnoreCase);
                    if (isSpecificFinder)
                    {
                        attrs = AddAttr("Default", !specificFinderFound ? "true" : "false", attrs);

                        List<NodeMatch> parameters = GetNodes("Parameter", "Direction", "Return", nodeMatch.Content);
                        bool isCollection = GetNodes("TypeDescriptor", "IsCollection", "true", parameters[0].Content).Count > 0;                        

                        if (isCollection)
                            attrs = AddAttr("ReturnTypeDescriptorPath", (returnTypeDescriptorNameVal + "[0]").Replace(".", @"\.\"), attrs);

                        specificFinderFound = true;
                    }

                    sb.Replace(methodInstanceNodeMatch.Attributes, " " + attrs);
                }
            }

            return sb.ToString();
        }

        protected string AdjustTypeDescriptorTags(string source)
        {
            string res = ReplaceNode("Entities", source, delegate(ReplaceNodeMatch replaceNodeMatch)
            {
                XmlDocument xmlDoc = new XmlDocument();
                xmlDoc.LoadXml(replaceNodeMatch.FoundNodeMatch.WholeNode);
                XmlNodeList nodeList  = xmlDoc.SelectNodes(@"//TypeDescriptor[contains(@TypeName, ""System.DateTime"")]");

                foreach (XmlNode node in nodeList)
                    node.AppendChild(CreateNode(xmlDoc, string.Format(@"<Interpretation><NormalizeDateTime LobDateTimeMode=""{0}"" /></Interpretation>", _lobDateTimeMode)));

                return xmlDoc.InnerXml;                
            });
            return res;
        }        

        protected static string RemoveAttr(string attrName, string source, out string attrVal)
        {
            attrVal = null;

            string attrValTmp = null;
            string res = Regex.Replace(source, string.Format(_attrExtractorTemplate, attrName), delegate(Match match)
            {
                attrValTmp = match.Groups["val"].Value;
                return string.Empty;
            }, RegexOptions.IgnoreCase | RegexOptions.Compiled);

            attrVal = attrValTmp;

            return res.Trim();
        }

        protected static string RemoveAttr(string attrName, string source)
        {
            string attrVal = null;
            return RemoveAttr(attrName, source, out attrVal);
        }

        protected static string GetAttrVal(string attrName, string source)
        {
            string res = null;

            Match attrMatch = Regex.Match(source, string.Format(_attrExtractorTemplate, attrName), RegexOptions.IgnoreCase | RegexOptions.Compiled);
            if (attrMatch.Success)
                res = attrMatch.Groups["val"].Value;

            return res;
        }

        protected static string AddAttr(string attrName, string attrValue, string source)
        {
            return source.TrimEnd() + " " + attrName + "=\"" + attrValue + "\"";
        }

        protected static NodeMatch CreateNodeMatch(Match match)
        {
            return new NodeMatch() 
            { 
                WholeNode       = match.Value, 
                Content         = match.Groups["content"].Value, 
                Attributes      = match.Groups["attr"].Value, 
                WholeOpeningTag = match.Groups["openTag"].Value, 
                TagName         = match.Groups["tagName"].Value,                 
            };
        }

        protected static List<NodeMatch> GetNodes(string nodeName, string source)
        {
            List<NodeMatch> res = new List<NodeMatch>();
            foreach (Match match in Regex.Matches(source, string.Format(_tagExtractorTemplate, nodeName), RegexOptions.IgnoreCase | RegexOptions.Compiled))
                res.Add(CreateNodeMatch(match));
            return res;
        }

        protected static List<NodeMatch> GetNodes(string nodeName, string attrName, string attrValue, string source)
        {
            List<NodeMatch> tmpRes = GetNodes(nodeName, source);
            List<NodeMatch> res    = new List<NodeMatch>(tmpRes.Count);
            foreach (NodeMatch nodeMatch in tmpRes)
            {
                string attrVal = GetAttrVal(attrName, nodeMatch.Attributes);
                if (attrVal != null && attrVal.Equals(attrValue, StringComparison.OrdinalIgnoreCase))
                    res.Add(nodeMatch);
            }
            return res;
        }

        protected static string ReplaceNode(string nodeName, string source, Func<ReplaceNodeMatch, string> evaluator)
        {
            return Regex.Replace(source, string.Format(_tagExtractorTemplate, nodeName), delegate(Match match)
                {
                    ReplaceNodeMatch replaceNodeMatch = new ReplaceNodeMatch() { FoundNodeMatch = CreateNodeMatch(match), OriginMatch = match };
                    string res = evaluator(replaceNodeMatch);
                    if (res == null)
                        res = string.Empty;
                    return replaceNodeMatch.Skip ? match.Value : res;
                }, RegexOptions.IgnoreCase | RegexOptions.Compiled);
        }

        protected static string FormNode(NodeMatch nodeMatch)
        {
            string attrs   = nodeMatch.Attributes != null ? nodeMatch.Attributes.Trim() : string.Empty;
            string content = nodeMatch.Content    != null ? nodeMatch.Content.Trim()    : string.Empty;
            string name    = nodeMatch.TagName.Trim();

            return content.Length != 0 ? 
                string.Format("<{0}{1}>{2}</{0}>", name, !string.IsNullOrEmpty(attrs) ? (" " + attrs) : string.Empty, content) :
                string.Format("<{0}{1} />", name, !string.IsNullOrEmpty(attrs) ? (" " + attrs) : string.Empty);
        }

        protected static XmlNode CreateNode(XmlDocument xmlDoc, string nodesXml)
        {
            XmlDocumentFragment xmlDocFragment = xmlDoc.CreateDocumentFragment();
            xmlDocFragment.InnerXml = nodesXml;
            return xmlDocFragment;
        }

        protected static string XmlDocToString(XmlDocument xmlDoc)
        {
            StringWriter sw = new StringWriter();
            XmlTextWriter xw = new XmlTextWriter(sw);
            xw.Formatting = Formatting.Indented;
            xmlDoc.WriteTo(xw);
            return sw.ToString();
        }
    }
}
